from ckantools.decorators import action

from ckanext.versioned_datastore.lib.utils import is_datastore_resource
from ckanext.versioned_datastore.logic.resource import helptext, schema


@action(schema.vds_resource_check(), helptext.vds_resource_check)
def vds_resource_check(resource_id: str) -> bool:
    """
    Checks if the given resource ID is a datastore resource. This is defined as whether
    it has data in Elasticsearch or not.

    :param resource_id: the resource ID
    :returns: True if the resource has data in Elasticsearch, False if not
    """
    return is_datastore_resource(resource_id)



from ckantools.validators import list_of_strings, list_validator

from ckanext.datastore.logic.schema import json_validator
from ckanext.versioned_datastore.logic.validators import (
    boolean_validator,
    float_validator,
    ignore_missing,
    int_validator,
    not_empty,
    validate_datastore_resource_ids,
)

multi_params = {
    'resource_ids': [ignore_missing, validate_datastore_resource_ids],
    'query': [ignore_missing, json_validator],
    'version': [ignore_missing, int_validator],
    'query_version': [ignore_missing, str],
}

multi_paging = {
    'after': [ignore_missing, list_validator],
    'size': [ignore_missing, int_validator],
}


def vds_multi_query() -> dict:
    return {**multi_params, **multi_paging}


def vds_multi_count() -> dict:
    return {**multi_params}


def vds_multi_autocomplete_value() -> dict:
    return {
        'field': [not_empty, str],
        'prefix': [ignore_missing, str],
        'after': [ignore_missing, str],
        'size': [ignore_missing, int_validator],
        **multi_params,
    }


def vds_multi_autocomplete_field_latest() -> dict:
    return {
        'resource_ids': [validate_datastore_resource_ids],
        'text': [ignore_missing, str],
        'lowercase': [ignore_missing, boolean_validator],
    }


def vds_multi_autocomplete_field() -> dict:
    return {
        'resource_ids': [validate_datastore_resource_ids],
        'text': [ignore_missing, str],
        'lowercase': [ignore_missing, boolean_validator],
        'version': [ignore_missing, int_validator],
    }


def vds_multi_hash() -> dict:
    return {
        'query': [json_validator],
        'query_version': [ignore_missing, str],
    }


def vds_multi_fields() -> dict:
    return {
        'resource_ids': [validate_datastore_resource_ids],
        'version': [ignore_missing, int_validator],
        'size': [ignore_missing, int_validator],
        'ignore_groups': [ignore_missing, list_of_strings()],
        'sample': [ignore_missing, float_validator],
    }


def vds_multi_stats() -> dict:
    return {
        **multi_params,
        'missing': [ignore_missing, float],
        'field': [not_empty, str],
    }


def vds_multi_direct() -> dict:
    return {
        'resource_ids': [validate_datastore_resource_ids],
        'search': [ignore_missing, json_validator],
        'version': [ignore_missing, str],
    }



# if the resource has been side loaded into the datastore then this should be its URL
DATASTORE_ONLY_RESOURCE = '_datastore_only_resource'
# the formats we support for ingestion
SV_FORMATS = {'csv', 'tsv'}
XLS_FORMATS = {'xls'}
XLSX_FORMATS = {'xlsx'}
ZIP_FORMATS = {'zip'}
ALL_FORMATS = SV_FORMATS | XLS_FORMATS | XLSX_FORMATS | ZIP_FORMATS
# version used for resources which aren't in the datastore
NON_DATASTORE_VERSION = -1



from ckanext.datastore.logic.schema import json_validator
from ckanext.versioned_datastore.logic.validators import ignore_missing, not_missing


def vds_schema_latest() -> dict:
    return {}


def vds_schema_validate() -> dict:
    return {
        'query': [not_missing, json_validator],
        'query_version': [ignore_missing, str],
    }



from ckantools.validators import list_of_strings, list_validator

from ckanext.datastore.logic.schema import json_validator, unicode_or_json_validator
from ckanext.versioned_datastore.logic.validators import (
    boolean_validator,
    ignore_missing,
    int_validator,
    not_empty,
    not_missing,
    validate_datastore_resource_id,
)

# common parameters used for searching a resource with a query and a version
basic_params = {
    'resource_id': [not_empty, str, validate_datastore_resource_id],
    'q': [ignore_missing, unicode_or_json_validator],
    'filters': [ignore_missing, json_validator],
    'version': [ignore_missing, int_validator],
}
# common parameters for paging
basic_paging = {
    'limit': [ignore_missing, int_validator],
    'offset': [ignore_missing, int_validator],
    'after': [ignore_missing, list_validator],
}


def vds_basic_query() -> dict:
    return {
        **basic_params,
        **basic_paging,
        'fields': [ignore_missing, list_of_strings()],
        'sort': [ignore_missing, list_of_strings()],
        'facets': [ignore_missing, list_of_strings()],
        'facet_limits': [ignore_missing, json_validator],
        'run_query': [ignore_missing, boolean_validator],
    }


def vds_basic_count() -> dict:
    return {**basic_params}


def vds_basic_autocomplete() -> dict:
    return {
        'field': [not_empty, str],
        'term': [not_missing, str],
        **basic_params,
        **basic_paging,
    }


def vds_basic_extent() -> dict:
    return {**basic_params}



from ckantools.decorators import auth


@auth('resource_update', {'resource_id': 'id'})
def vds_data_add(context, data_dict) -> dict:
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def vds_data_delete(context, data_dict) -> dict:
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def vds_data_sync(context, data_dict) -> dict:
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def record_show(context, data_dict) -> dict:
    return {'success': True}


@auth('resource_show', {'resource_id': 'id'}, anon=True)
def vds_data_get(context, data_dict) -> dict:
    return {'success': True}



vds_basic_query = """
This action allows you to search data in a resource. It is designed to function in a
similar way to CKAN's core datastore_search but with a few extra bells and whistles,
most prominently versioning. This allows the resource to be searched at any moment in
it's lifespan and have the data as it looked at that moment returned, even if it has
changed since.

If the resource to be searched is private then appropriate authorization is required.

Note that in the parameters listed below spaces should only be included if part of a
field name, so, for example, don't include any spaces in comma separated lists unless
needed.

:param resource_id: ID of the resource to be searched against
:type resource_id: string
:param q: full text query. If a string is passed, all fields are searched with the
          value. If a dict is passed each of the fields and values contained within will
          be searched as required (e.g. {"field1": "a", "field2": "b"}).
:type q: string or dictionary
:param filters: a dictionary of conditions that must be met to match a record
                (e.g {"field1": "a", "field2": "b"}) (optional)
:type filters: dictionary
:param after: search_after value for elasticsearch to paginate from (optional). Use this
              mechanism to do deep (beyond 10000 values) pagination. The values have to
              match the sort currently in use and therefore it's recommended that this
              value is not built but rather passed from the previous result's 'after'
              key.
:type after: a list of values
:param limit: maximum number of records to return (optional, default: 100)
:type limit: int
:param offset: offset this number of records (optional)
:type offset: int
:param fields: fields to return for each record (optional, default: all fields are
               returned)
:type fields: list or comma separated string
:param sort: list of field names with ordering. Ordering is ascending by default, if
             descending is required, add "desc" after the field name
             e.g.: "fieldname1,fieldname2 desc" sorts by fieldname1 asc and fieldname2
             desc
:type sort: list or comma separated string
:param version: version to search at, if not provided the current version of the data is
               searched.
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param facets: if present, the top 10 most frequent values for each of the fields in
               this list will be returned along with estimated counts for each value.
               Calculating these results has a reasonable overhead so only include this
               parameter if you need it
:type facets: list or comma separated string
:param facet_limits: if present, specifies the number of top values to retrieve for the
                    facets listed within. The default number will be used if this
                    parameter is not specified or if a facet in the facets list does not
                    appear in this dict. For example, with this facet list
                    ['facet1', 'facet2', 'facet3', 'facet4'], and this facet_limits
                    dict: {'facet1': 50, 'facet4': 10}, facet1 and facet4 would be
                    limited to top 50 and 10 values respectively, whereas facet2 and
                    facet3 would be limited to the default of the top 10.
:type facet_limits: a dict
:param run_query: boolean value indicating whether the query should be run and the
                  results returned or whether the query should be created and the
                  elasticsearch query returned instead of the results. Defaults to True.
:type run_query: boolean


**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param fields: fields/columns and their extra metadata
:type fields: list of dicts
:param total: number of total matching records
:type total: int
:param records: list of matching results
:type records: list of dicts
:param facets: list of fields and their top 10 values, if requested
:type facets: dict
:param after: the next page's search_after value which can be passed back as the "after"
              parameter. This value will always be included if there were results
              otherwise None is returned. A value will also always be returned even if
              this page is the last.
:type after: a list or None

If run_query is True, then a dict with the following keys is returned instead:

:param indexes: a list of the fully qualified indexes that the query would have been run
                against
:type indexes: a list of strings
:param search: the query dict that would have been sent to elasticsearch
:type search: dict
"""

vds_basic_count = ''
vds_basic_autocomplete = ''
vds_basic_extent = ''
vds_basic_raw = ''



import logging
from contextlib import suppress
from typing import List, Optional

from ckan.plugins import (
    SingletonPlugin,
    implements,
    interfaces,
    toolkit,
)
from ckantools.loaders import create_actions, create_auth
from elasticsearch import Elasticsearch
from pymongo import MongoClient
from splitgill.manager import SplitgillClient

from ckanext.versioned_datastore import cli, helpers, routes
from ckanext.versioned_datastore.interfaces import IVersionedDatastoreQuerySchema
from ckanext.versioned_datastore.lib.query.schema import register_schema
from ckanext.versioned_datastore.lib.query.schemas.v1_0_0 import v1_0_0Schema
from ckanext.versioned_datastore.lib.query.search.query import SchemaQuery
from ckanext.versioned_datastore.lib.tasks import get_es_health, get_queue_length
from ckanext.versioned_datastore.lib.utils import (
    RawResourceException,
    ReadOnlyResourceException,
    iqs_implementations,
    is_datastore_resource,
    ivds_implementations,
)
from ckanext.versioned_datastore.logic.basic import (
    action as basic_action,
)
from ckanext.versioned_datastore.logic.basic import (
    auth as basic_auth,
)
from ckanext.versioned_datastore.logic.data import (
    action as data_action,
)
from ckanext.versioned_datastore.logic.data import (
    auth as data_auth,
)
from ckanext.versioned_datastore.logic.download import (
    action as download_action,
)
from ckanext.versioned_datastore.logic.download import (
    auth as download_auth,
)
from ckanext.versioned_datastore.logic.multi import (
    action as multi_action,
)
from ckanext.versioned_datastore.logic.multi import (
    auth as multi_auth,
)
from ckanext.versioned_datastore.logic.options import (
    action as options_action,
)
from ckanext.versioned_datastore.logic.options import (
    auth as options_auth,
)
from ckanext.versioned_datastore.logic.resource import (
    action as resource_action,
)
from ckanext.versioned_datastore.logic.resource import (
    auth as resource_auth,
)
from ckanext.versioned_datastore.logic.schema import (
    action as schema_action,
)
from ckanext.versioned_datastore.logic.schema import (
    auth as schema_auth,
)
from ckanext.versioned_datastore.logic.slug import (
    action as slug_action,
)
from ckanext.versioned_datastore.logic.slug import (
    auth as slug_auth,
)
from ckanext.versioned_datastore.logic.version import (
    action as version_action,
)
from ckanext.versioned_datastore.logic.version import (
    auth as version_auth,
)

try:
    from ckanext.status.interfaces import IStatus

    status_available = True
except ImportError:
    status_available = False

log = logging.getLogger(__name__)

# stop elasticsearch from showing warning logs
logging.getLogger('elasticsearch').setLevel(logging.ERROR)


class VersionedSearchPlugin(SingletonPlugin):
    implements(interfaces.IActions)
    implements(interfaces.IAuthFunctions)
    implements(interfaces.ITemplateHelpers, inherit=True)
    implements(interfaces.IResourceController, inherit=True)
    implements(interfaces.IConfigurer)
    implements(interfaces.IConfigurable)
    implements(interfaces.IBlueprint, inherit=True)
    implements(IVersionedDatastoreQuerySchema)
    implements(interfaces.IClick)
    if status_available:
        implements(IStatus)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # all these are set up during the configure method
        self.mongo_client: Optional[MongoClient] = None
        self.elasticsearch_client: Optional[Elasticsearch] = None
        self.sg_client: Optional[SplitgillClient] = None

    # IConfigurable
    def configure(self, ckan_config):
        # do the setup for Splitgill first, create Mongo client, Elasticsearch client,
        # and then the Splitgill client
        self.mongo_client = MongoClient(
            host=ckan_config.get('ckanext.versioned_datastore.mongo_host'),
            port=int(ckan_config.get('ckanext.versioned_datastore.mongo_port')),
        )
        es_hosts = ckan_config.get(
            'ckanext.versioned_datastore.elasticsearch_hosts'
        ).split(',')
        es_port = ckan_config.get('ckanext.versioned_datastore.elasticsearch_port')
        self.elasticsearch_client = Elasticsearch(
            hosts=[f'http://{host}:{es_port}/' for host in es_hosts],
            # todo: check these params
            sniff_on_start=True,
            sniff_on_node_failure=True,
            sniff_timeout=30,
            http_compress=False,
            request_timeout=60,
        )
        mongo_db_name = ckan_config.get('ckanext.versioned_datastore.mongo_database')
        self.sg_client = SplitgillClient(
            self.mongo_client, self.elasticsearch_client, mongo_db_name
        )

        # register all custom query schemas
        for plugin in iqs_implementations():
            for query_version, query_schema in plugin.get_query_schemas():
                register_schema(query_version, query_schema)

        # reserve any requested slugs
        from .lib.query.slugs.slugs import reserve_slug

        for plugin in ivds_implementations():
            slugs = plugin.vds_reserve_slugs()
            for reserved_pretty_slug, query_parameters in slugs.items():
                query = SchemaQuery(**query_parameters)
                with suppress(Exception):
                    reserve_slug(reserved_pretty_slug, query)

    def is_sg_configured(self) -> bool:
        """
        Returns whether Splitgill is configured and ready for use. This checks if the
        Mongo client, Elasticsearch client, and Splitgill client have all been created
        and stored against this plugin instance.

        :returns: True if it's ready, False if not
        """
        return (
            self.mongo_client is not None
            and self.elasticsearch_client is not None
            and self.sg_client is not None
        )

    # IActions
    def get_actions(self):
        return create_actions(
            basic_action,
            data_action,
            download_action,
            multi_action,
            options_action,
            resource_action,
            schema_action,
            slug_action,
            version_action,
        )

    # IAuthFunctions
    def get_auth_functions(self):
        return create_auth(
            basic_auth,
            data_auth,
            download_auth,
            multi_auth,
            options_auth,
            resource_auth,
            schema_auth,
            slug_auth,
            version_auth,
        )

    # IClick
    def get_commands(self):
        return cli.get_commands()

    # ITemplateHelpers
    def get_helpers(self):
        return {
            'is_datastore_resource': is_datastore_resource,
            'is_duplicate_ingestion': helpers.is_duplicate_ingestion,
            'get_human_duration': helpers.get_human_duration,
            'get_stat_icon': helpers.get_stat_icon,
            'get_stat_activity_class': helpers.get_stat_activity_class,
            'get_stat_title': helpers.get_stat_title,
            'get_available_formats': helpers.get_available_formats,
            'pretty_print_json': helpers.pretty_print_json,
            'get_version_date': helpers.get_version_date,
            'latest_item_version': helpers.latest_item_version,
            'nav_slug': helpers.nav_slug,
        }

    # IResourceController
    def before_show(self, resource_dict):
        # ensure datastore_active is set where it should be
        resource_dict['datastore_active'] = is_datastore_resource(resource_dict['id'])
        # theoretically a resource could be datastore_active and have parsing disabled
        # at the same time if the database and ES have gotten out of sync, which isn't
        # ideal, but the fixes are more annoying than the problem itself
        return resource_dict

    # IResourceController
    def before_create(self, context, resource):
        # only set disable_parsing if it's True (False by default)
        if toolkit.asbool(resource.pop('disable_parsing', False)):
            resource['disable_parsing'] = True

    # IResourceController
    def before_update(self, context, current, resource):
        # we can't automatically go from ingested to raw because it might be included
        # in queries and DOIs, but we can ingest a file that was previously raw
        was_raw = toolkit.asbool(current.get('disable_parsing', False))
        is_raw = toolkit.asbool(resource.pop('disable_parsing', False))
        if was_raw and is_raw:
            # only allow if it was already previously raw
            resource['disable_parsing'] = True

    # IResourceController
    def after_update(self, context: dict, resource: dict):
        # use replace to overwrite the existing data (this is what users would expect)
        data_dict = {'resource_id': resource['id'], 'replace': True}
        with suppress(ReadOnlyResourceException), suppress(RawResourceException):
            toolkit.get_action('vds_data_add')(context, data_dict)

    # IResourceController
    def after_create(self, context: dict, resource: dict):
        # use replace to overwrite the existing data (this is what users would expect)
        data_dict = {'resource_id': resource['id'], 'replace': True}
        with suppress(ReadOnlyResourceException), suppress(RawResourceException):
            toolkit.get_action('vds_data_add')(context, data_dict)

    def before_delete(self, context: dict, resource: dict, resources: List[dict]):
        toolkit.get_action('vds_data_delete')(context, {'resource_id': resource['id']})

    # IConfigurer
    def update_config(self, config):
        # add public folder containing schemas
        toolkit.add_public_directory(config, 'theme/public')
        # add templates
        toolkit.add_template_directory(config, 'theme/templates')
        toolkit.add_resource('theme/assets', 'ckanext-versioned-datastore')

    # IBlueprint
    def get_blueprint(self):
        return routes.blueprints

    # IVersionedDatastoreQuerySchema
    def get_query_schemas(self):
        return [(v1_0_0Schema.version, v1_0_0Schema())]

    # IStatus
    def modify_status_reports(self, status_reports):
        queued_downloads = get_queue_length('download')

        status_reports.append(
            {
                'label': toolkit._('Downloads'),
                'value': queued_downloads,
                'group': toolkit._('Queues'),
                'help': toolkit._('Number of downloads waiting in the queue'),
                'state': 'good'
                if queued_downloads < 2
                else ('ok' if queued_downloads < 4 else 'bad'),
            }
        )

        queued_imports = get_queue_length('importing')

        status_reports.append(
            {
                'label': toolkit._('Imports'),
                'value': queued_imports,
                'group': toolkit._('Queues'),
                'help': toolkit._('Number of import jobs waiting in the queue'),
                'state': 'good'
                if queued_imports < 2
                else ('ok' if queued_imports < 4 else 'bad'),
            }
        )

        es_health = get_es_health()
        server_status_text = (
            toolkit._('available') if es_health['ping'] else toolkit._('unavailable')
        )

        status_reports.append(
            {
                'label': toolkit._('Search'),
                'value': server_status_text,
                'help': toolkit._(
                    'Multisearch functionality is provided by an Elasticsearch cluster'
                ),
                'state': 'good' if es_health['ping'] else 'bad',
            }
        )

        return status_reports



from datetime import datetime as dt
from unittest.mock import MagicMock, patch

import pytest
from ckan.plugins import toolkit

from ckanext.versioned_datastore.logic.slug.action import vds_slug_resolve

warnings = {
    'NO_RESOURCES': 'Resource IDs were not saved for this search. The list of '
    'available resources may have changed since this slug was saved.',
    'ALL_INVALID_RESOURCES': 'All resources associated with this search have been '
    'deleted, moved, or are no longer available.',
    'SOME_INVALID_RESOURCES': 'Some resources have been deleted, moved, or are no '
    'longer available. Affected resources: {}',
    'RECORD_COUNT': 'Record count at save time: {}',
}


@pytest.mark.usefixtures('with_request_context', 'with_vds', 'with_vds_resource')
class TestResolveSlug:
    @patch('ckanext.versioned_datastore.logic.slug.action.resolve_slug')
    def test_resolve_slug_basic(self, mock_resolve_slug, with_vds_resource):
        """
        Test resolving a regular vds slug with no modifications or warnings.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        query = {'search': 'banana'}
        mock_resolve_slug.return_value = MagicMock(
            query=query,
            query_version='v1.0.0',
            resource_ids=[
                resource_1['id'],
                resource_2['id'],
            ],
            version=1234,
            created=ts,
        )
        action_response = vds_slug_resolve('test-slug')

        assert action_response['query'] == query
        assert action_response['query_version'] == 'v1.0.0'
        assert resource_1['id'] in action_response['resource_ids']
        assert resource_2['id'] in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 2
        assert action_response['version'] == 1234
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 0

    @patch('ckanext.versioned_datastore.logic.slug.action.resolve_slug')
    def test_resolve_slug_v0(self, mock_resolve_slug, with_vds_resource):
        """
        Test resolving a v0 vds slug.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        v0_query = {'filters': {'_id': ['1']}}
        v1_query = {
            'filters': {'and': [{'string_equals': {'fields': ['_id'], 'value': '1'}}]}
        }
        mock_resolve_slug.return_value = MagicMock(
            query=v0_query,
            query_version='v0',
            resource_ids=[
                resource_1['id'],
                resource_2['id'],
            ],
            version=1234,
            created=ts,
        )
        action_response = vds_slug_resolve('test-slug')

        assert action_response['query'] == v1_query
        assert action_response['query_version'] == 'v1.0.0'  # current version
        assert resource_1['id'] in action_response['resource_ids']
        assert resource_2['id'] in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 2
        assert action_response['version'] == 1234
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 0

    @patch('ckanext.versioned_datastore.logic.slug.action.resolve_slug')
    def test_resolve_slug_no_resources(self, mock_resolve_slug, with_vds_resource):
        """
        Test resolving a vds slug with no resources.
        """
        ts = dt(2025, 1, 1)
        mock_resolve_slug.return_value = MagicMock(
            query={},
            query_version='v1.0.0',
            resource_ids=[],
            version=1234,
            created=ts,
        )
        action_response = vds_slug_resolve('test-slug')

        assert action_response['query'] == {}
        assert action_response['query_version'] == 'v1.0.0'
        assert len(action_response['resource_ids']) == 0
        assert action_response['version'] == 1234
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 1
        assert warnings['NO_RESOURCES'] in action_response['warnings']

    @patch('ckanext.versioned_datastore.logic.slug.action.resolve_slug')
    def test_resolve_slug_some_invalid_resources(
        self, mock_resolve_slug, with_vds_resource
    ):
        """
        Test resolving a vds slug where some but not all of the resources are invalid.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        mock_resolve_slug.return_value = MagicMock(
            query={},
            query_version='v1.0.0',
            resource_ids=[
                resource_1['id'],
                'invalid-resource-id',
            ],
            version=1234,
            created=ts,
        )
        action_response = vds_slug_resolve('test-slug')

        assert action_response['query'] == {}
        assert action_response['query_version'] == 'v1.0.0'
        assert resource_1['id'] in action_response['resource_ids']
        assert 'invalid-resource-id' not in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 1
        assert action_response['version'] == 1234
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 1
        assert (
            warnings['SOME_INVALID_RESOURCES'].format(str(1))
            in action_response['warnings']
        )

    @patch('ckanext.versioned_datastore.logic.slug.action.resolve_slug')
    def test_resolve_slug_all_invalid_resources(self, mock_resolve_slug):
        """
        Test resolving a vds slug where all of the resources are invalid.
        """
        ts = dt(2025, 1, 1)
        mock_resolve_slug.return_value = MagicMock(
            query={},
            query_version='v1.0.0',
            resource_ids=[
                'invalid-resource-id',
            ],
            version=1234,
            created=ts,
        )
        with pytest.raises(
            toolkit.ValidationError, match=warnings['ALL_INVALID_RESOURCES']
        ):
            vds_slug_resolve('test-slug')

    @patch(
        'ckanext.versioned_datastore.logic.slug.action.plugin_loaded',
        return_value=False,
    )
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.resolve_slug', return_value=None
    )
    def test_no_slug(self, *args):
        """
        Test trying to resolve a string that is neither a slug nor a DOI.
        """
        with pytest.raises(
            toolkit.ValidationError, match='This saved search could not be found'
        ):
            vds_slug_resolve('slug-does-not-exist')


@pytest.mark.usefixtures('with_request_context', 'with_vds', 'with_vds_resource')
class TestResolveDOI:
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.plugin_loaded', return_value=True
    )
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.resolve_slug', return_value=None
    )
    @patch('ckanext.query_dois.model.QueryDOI')  # we're not testing this
    def test_resolve_doi_basic(
        self, mock_plugin_loaded, mock_resolve_slug, mock_query_doi, with_vds_resource
    ):
        """
        Test resolving a query DOI.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        query = {'search': 'banana'}
        mock_resolved = MagicMock(
            query=query,
            query_version='v1.0.0',
            requested_version=1357,
            timestamp=ts,
            count=1000,
        )
        mock_resolved.get_rounded_versions.return_value = [1234, 5678]
        mock_resolved.get_resource_ids.return_value = [
            resource_1['id'],
            resource_2['id'],
        ]
        with patch('ckan.model.Session') as mock_session:
            mock_session.query.return_value.filter.return_value.first.return_value = (
                mock_resolved
            )
            action_response = vds_slug_resolve('test-doi')

        assert action_response['query'] == query
        assert action_response['query_version'] == 'v1.0.0'
        assert resource_1['id'] in action_response['resource_ids']
        assert resource_2['id'] in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 2
        assert action_response['version'] == 1357
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 0

    @patch(
        'ckanext.versioned_datastore.logic.slug.action.plugin_loaded', return_value=True
    )
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.resolve_slug', return_value=None
    )
    @patch('ckanext.query_dois.model.QueryDOI')  # we're not testing this
    def test_resolve_doi_no_requested_version(
        self, mock_plugin_loaded, mock_resolve_slug, mock_query_doi, with_vds_resource
    ):
        """
        Test resolving a query DOI where the requested_version has not been specified.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        query = {'search': 'banana'}
        mock_resolved = MagicMock(
            query=query,
            query_version='v1.0.0',
            requested_version=None,
            timestamp=ts,
            count=1000,
        )
        mock_resolved.get_rounded_versions.return_value = [1234, 5678]
        mock_resolved.get_resource_ids.return_value = [
            resource_1['id'],
            resource_2['id'],
        ]
        with patch('ckan.model.Session') as mock_session:
            mock_session.query.return_value.filter.return_value.first.return_value = (
                mock_resolved
            )
            action_response = vds_slug_resolve('test-doi')

        assert action_response['query'] == query
        assert action_response['query_version'] == 'v1.0.0'
        assert resource_1['id'] in action_response['resource_ids']
        assert resource_2['id'] in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 2
        assert action_response['version'] == 5678
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 0

    @patch(
        'ckanext.versioned_datastore.logic.slug.action.plugin_loaded', return_value=True
    )
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.resolve_slug', return_value=None
    )
    @patch('ckanext.query_dois.model.QueryDOI')  # we're not testing this
    def test_resolve_doi_v0(
        self, mock_plugin_loaded, mock_resolve_slug, mock_query_doi, with_vds_resource
    ):
        """
        Test resolving a v0 query DOI.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        v0_query = {'filters': {'_id': ['1']}}
        v1_query = {
            'filters': {'and': [{'string_equals': {'fields': ['_id'], 'value': '1'}}]}
        }
        mock_resolved = MagicMock(
            query=v0_query,
            query_version='v0',
            requested_version=1357,
            timestamp=ts,
            count=1000,
        )
        mock_resolved.get_rounded_versions.return_value = [1234, 5678]
        mock_resolved.get_resource_ids.return_value = [
            resource_1['id'],
            resource_2['id'],
        ]
        with patch('ckan.model.Session') as mock_session:
            mock_session.query.return_value.filter.return_value.first.return_value = (
                mock_resolved
            )
            action_response = vds_slug_resolve('test-doi')

        assert action_response['query'] == v1_query
        assert action_response['query_version'] == 'v1.0.0'
        assert resource_1['id'] in action_response['resource_ids']
        assert resource_2['id'] in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 2
        assert action_response['version'] == 1357
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 0

    @patch(
        'ckanext.versioned_datastore.logic.slug.action.plugin_loaded', return_value=True
    )
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.resolve_slug', return_value=None
    )
    @patch('ckanext.query_dois.model.QueryDOI')  # we're not testing this
    def test_resolve_doi_some_invalid_resources(
        self, mock_plugin_loaded, mock_resolve_slug, mock_query_doi, with_vds_resource
    ):
        """
        Test resolving a query DOI with some invalid resources.
        """
        resource_1, resource_2 = with_vds_resource
        ts = dt(2025, 1, 1)
        mock_resolved = MagicMock(
            query={},
            query_version='v1.0.0',
            requested_version=1357,
            timestamp=ts,
            count=1000,
        )
        mock_resolved.get_rounded_versions.return_value = [1234, 5678]
        mock_resolved.get_resource_ids.return_value = [
            resource_1['id'],
            'invalid-resource-id',
        ]
        with patch('ckan.model.Session') as mock_session:
            mock_session.query.return_value.filter.return_value.first.return_value = (
                mock_resolved
            )
            action_response = vds_slug_resolve('test-doi')

        assert action_response['query'] == {}
        assert action_response['query_version'] == 'v1.0.0'
        assert resource_1['id'] in action_response['resource_ids']
        assert 'invalid-resource-id' not in action_response['resource_ids']
        assert len(action_response['resource_ids']) == 1
        assert action_response['version'] == 1357
        assert action_response['created'] == '2025-01-01T00:00:00'
        assert len(action_response['warnings']) == 2
        assert (
            warnings['SOME_INVALID_RESOURCES'].format(str(1))
            in action_response['warnings']
        )
        assert warnings['RECORD_COUNT'].format(str(1000)) in action_response['warnings']

    @patch(
        'ckanext.versioned_datastore.logic.slug.action.plugin_loaded', return_value=True
    )
    @patch(
        'ckanext.versioned_datastore.logic.slug.action.resolve_slug', return_value=None
    )
    @patch('ckanext.query_dois.model.QueryDOI')  # we're not testing this
    def test_resolve_doi_all_invalid_resources(
        self, mock_plugin_loaded, mock_resolve_slug, mock_query_doi, with_vds_resource
    ):
        """
        Test resolving a query DOI with no valid resources.
        """
        ts = dt(2025, 1, 1)
        mock_resolved = MagicMock(
            query={},
            query_version='v1.0.0',
            requested_version=1357,
            timestamp=ts,
            count=1000,
        )
        mock_resolved.get_rounded_versions.return_value = [1234, 5678]
        mock_resolved.get_resource_ids.return_value = [
            'invalid-resource-id',
        ]
        with patch('ckan.model.Session') as mock_session:
            mock_session.query.return_value.filter.return_value.first.return_value = (
                mock_resolved
            )
            with pytest.raises(
                toolkit.ValidationError, match=warnings['ALL_INVALID_RESOURCES']
            ):
                vds_slug_resolve('test-doi')



import json
from unittest.mock import MagicMock

import pytest
from ckan.plugins import toolkit
from elasticsearch_dsl.query import EMPTY_QUERY, Bool, GeoShape
from splitgill.search import ALL_POINTS, ALL_TEXT, keyword, text

from ckanext.versioned_datastore.lib.query.search.query import BasicQuery


class TestAddPointFilter:
    def test_simple(self):
        distance = '4km'
        coordinates = [4.3, 1.2]

        query = BasicQuery.create_point_filter(distance, coordinates)

        assert query.to_dict() == {
            'geo_distance': {
                'distance': '4km',
                ALL_POINTS: {
                    'lat': coordinates[1],
                    'lon': coordinates[0],
                },
            }
        }

    def test_float_conversion(self):
        distance = '4km'
        coordinates = ['4.53423', 100]

        query = BasicQuery.create_point_filter(distance, coordinates)

        assert query.to_dict() == {
            'geo_distance': {
                'distance': '4km',
                ALL_POINTS: {
                    'lat': 100.0,
                    'lon': 4.53423,
                },
            }
        }


class TestAddMultiPolygonFilter:
    def test_single(self):
        coordinates = [[[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]]]

        query = BasicQuery.create_multipolygon_filter(coordinates)

        assert query == GeoShape(
            **{
                ALL_POINTS: {
                    'shape': {
                        'type': 'Polygon',
                        'coordinates': [
                            [[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]
                        ],
                    }
                }
            }
        )

    def test_multi(self):
        coordinates = [
            [[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]],
            [[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]],
        ]

        query = BasicQuery.create_multipolygon_filter(coordinates)

        expected = GeoShape(
            **{
                ALL_POINTS: {
                    'shape': {
                        'type': 'Polygon',
                        'coordinates': [
                            [
                                [-16, 44],
                                [-13.1, 34.8],
                                [15.99, 35],
                                [5, 49],
                            ]
                        ],
                    }
                }
            }
        )
        assert query == Bool(should=[expected, expected], minimum_should_match=1)

    def test_validation_error(self):
        # only two points!
        coordinates = [[[[-16, 44], [-13.1, 34.8]]]]

        with pytest.raises(toolkit.ValidationError):
            BasicQuery.create_multipolygon_filter(coordinates)

    def test_float_conversion(self):
        coordinates = [
            [[['-16', '44'], ['-13.1', '34.8'], ['15.99', '35'], ['5', '49']]]
        ]

        query = BasicQuery.create_multipolygon_filter(coordinates)

        assert query == GeoShape(
            **{
                ALL_POINTS: {
                    'shape': {
                        'type': 'Polygon',
                        'coordinates': [
                            [
                                [-16, 44],
                                [-13.1, 34.8],
                                [15.99, 35],
                                [5, 49],
                            ]
                        ],
                    },
                }
            }
        )


class TestAddPolygonFilter:
    def test_simple(self):
        coordinates = [[['-16', '44'], ['-13.1', '34.8'], ['15.99', '35'], ['5', '49']]]

        query = BasicQuery.create_polygon_filter(coordinates)

        assert query == GeoShape(
            **{
                ALL_POINTS: {
                    'shape': {
                        'type': 'Polygon',
                        'coordinates': [
                            [
                                [-16, 44],
                                [-13.1, 34.8],
                                [15.99, 35],
                                [5, 49],
                            ]
                        ],
                    }
                }
            }
        )


class TestAddGeoSearch:
    def test_valid_point_dict(self):
        geo_filter = {
            'type': 'Point',
            'distance': MagicMock(),
            'coordinates': MagicMock(),
        }
        query = BasicQuery.create_geo_filter(geo_filter)
        assert query == BasicQuery.create_point_filter(
            geo_filter['distance'],
            geo_filter['coordinates'],
        )

    def test_valid_point_string(self):
        geo_filter = json.dumps(
            {
                'type': 'Point',
                'distance': '10km',
                'coordinates': [40, 30],
            }
        )
        query = BasicQuery.create_geo_filter(geo_filter)
        assert query == BasicQuery.create_point_filter(
            '10km',
            [40, 30],
        )

    def test_invalid_point(self):
        with pytest.raises(toolkit.ValidationError):
            geo_filter = {
                'type': 'Point',
            }
            BasicQuery.create_geo_filter(geo_filter)

        for param in ('distance', 'coordinates'):
            with pytest.raises(toolkit.ValidationError):
                geo_filter = {
                    'type': 'Point',
                    param: MagicMock(),
                }
                BasicQuery.create_geo_filter(geo_filter)

    def test_valid_multipolygon_dict(self):
        geo_filter = {
            'type': 'MultiPolygon',
            'coordinates': [[[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]]],
        }
        query = BasicQuery.create_geo_filter(geo_filter)
        assert query == BasicQuery.create_multipolygon_filter(geo_filter['coordinates'])

    def test_valid_multipolygon_string(self):
        coords = [[[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]]]
        geo_filter = json.dumps({'type': 'MultiPolygon', 'coordinates': coords})
        query = BasicQuery.create_geo_filter(geo_filter)
        assert query == BasicQuery.create_multipolygon_filter(coords)

    def test_invalid_multipolygon(self):
        with pytest.raises(toolkit.ValidationError):
            geo_filter = {
                'type': 'MultiPolygon',
            }
            BasicQuery.create_geo_filter(geo_filter)

    def test_valid_polygon_dict(self):
        geo_filter = {
            'type': 'Polygon',
            'coordinates': [[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]],
        }
        query = BasicQuery.create_geo_filter(geo_filter)
        assert query == BasicQuery.create_polygon_filter(geo_filter['coordinates'])

    def test_valid_polygon_string(self):
        geo_filter = {
            'type': 'Polygon',
            'coordinates': [[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]],
        }
        query = BasicQuery.create_geo_filter(json.dumps(geo_filter))
        assert query == BasicQuery.create_polygon_filter(geo_filter['coordinates'])

    def test_invalid_polygon(self):
        with pytest.raises(toolkit.ValidationError):
            geo_filter = {
                'type': 'Polygon',
            }
            BasicQuery.create_geo_filter(geo_filter)

    def test_invalid_type(self):
        with pytest.raises(toolkit.ValidationError):
            # a type we don't support
            BasicQuery.create_geo_filter({'type': 'not a type we support'})

        with pytest.raises(toolkit.ValidationError):
            # no type
            BasicQuery.create_geo_filter({})

    def test_invalid_json(self):
        # here's a valid geo_filter value
        geo_filter = {
            'type': 'Polygon',
            # can't use magic mocks here because they need to be JSON serialised
            'coordinates': 'coord mock',
        }
        # but we're gonna add the word beans at the end to make it invalid
        geo_filter = json.dumps(geo_filter) + 'beans'

        with pytest.raises(toolkit.ValidationError):
            BasicQuery.create_geo_filter(geo_filter)


class TestToDSL:
    def test_empty_q(self):
        assert BasicQuery('test').to_dsl() == EMPTY_QUERY
        assert BasicQuery('test', q='').to_dsl() == EMPTY_QUERY
        assert BasicQuery('test', q=None).to_dsl() == EMPTY_QUERY
        assert BasicQuery('test', q={}).to_dsl() == EMPTY_QUERY

    def _run_test(self, data_dict: dict, expected: dict):
        assert BasicQuery('test', **data_dict).to_dsl().to_dict() == expected

    def test_q_simple_text(self):
        self._run_test(
            {'q': 'banana'},
            {
                'match': {
                    ALL_TEXT: {
                        'query': 'banana',
                        'operator': 'and',
                    }
                }
            },
        )
        self._run_test(
            {'q': 'a multi-word example'},
            {
                'match': {
                    ALL_TEXT: {
                        'query': 'a multi-word example',
                        'operator': 'and',
                    }
                }
            },
        )

    def test_q_dicts(self):
        self._run_test(
            {'q': {'': 'banana'}},
            {
                'match': {
                    ALL_TEXT: {
                        'query': 'banana',
                        'operator': 'and',
                    }
                }
            },
        )
        self._run_test(
            {'q': {'field1': 'banana'}},
            {
                'match': {
                    text('field1'): {
                        'query': 'banana',
                        'operator': 'and',
                    }
                }
            },
        )
        self._run_test(
            {'q': {'field1': 'banana', 'field2': 'lemons'}},
            {
                'bool': {
                    'must': [
                        {
                            'match': {
                                text('field1'): {
                                    'query': 'banana',
                                    'operator': 'and',
                                },
                            }
                        },
                        {
                            'match': {
                                text('field2'): {
                                    'query': 'lemons',
                                    'operator': 'and',
                                },
                            }
                        },
                    ]
                }
            },
        )

    def test_q_string_and_unicode(self):
        self._run_test(
            {'q': 'a string'},
            {
                'match': {
                    ALL_TEXT: {
                        'query': 'a string',
                        'operator': 'and',
                    }
                }
            },
        )
        self._run_test(
            {'q': 'a unicode string'},
            {
                'match': {
                    ALL_TEXT: {
                        'query': 'a unicode string',
                        'operator': 'and',
                    }
                }
            },
        )

    def test_q_non_string(self):
        self._run_test(
            {'q': 4},
            {
                'match': {
                    ALL_TEXT: {
                        'query': '4',
                        'operator': 'and',
                    }
                }
            },
        )
        self._run_test(
            {'q': 4.31932},
            {
                'match': {
                    ALL_TEXT: {
                        'query': '4.31932',
                        'operator': 'and',
                    }
                }
            },
        )

    def test_filters_empties(self):
        self._run_test({'filters': {}}, EMPTY_QUERY.to_dict())

    def test_filters_non_lists(self):
        self._run_test(
            {
                'filters': {
                    'field1': 'banana',
                }
            },
            {
                'term': {
                    keyword('field1'): 'banana',
                }
            },
        )
        self._run_test(
            {
                'filters': {
                    'field1': 'banana',
                    'field2': 'lemons',
                }
            },
            {
                'bool': {
                    'filter': [
                        {
                            'term': {
                                keyword('field1'): 'banana',
                            }
                        },
                        {
                            'term': {
                                keyword('field2'): 'lemons',
                            }
                        },
                    ]
                }
            },
        )

    def test_filters_lists(self):
        self._run_test(
            {
                'filters': {
                    'field1': ['banana'],
                }
            },
            {
                'term': {
                    keyword('field1'): 'banana',
                }
            },
        )
        self._run_test(
            {
                'filters': {
                    'field1': ['banana'],
                    'field2': ['lemons'],
                }
            },
            {
                'bool': {
                    'filter': [
                        {
                            'term': {
                                keyword('field1'): 'banana',
                            }
                        },
                        {
                            'term': {
                                keyword('field2'): 'lemons',
                            }
                        },
                    ]
                }
            },
        )
        self._run_test(
            {
                'filters': {
                    'field1': ['banana', 'goat', 'funnel'],
                    'field2': ['lemons', 'chunk'],
                }
            },
            {
                'bool': {
                    'filter': [
                        {
                            'term': {
                                keyword('field1'): 'banana',
                            }
                        },
                        {
                            'term': {
                                keyword('field1'): 'goat',
                            }
                        },
                        {
                            'term': {
                                keyword('field1'): 'funnel',
                            }
                        },
                        {
                            'term': {
                                keyword('field2'): 'lemons',
                            }
                        },
                        {
                            'term': {
                                keyword('field2'): 'chunk',
                            }
                        },
                    ]
                }
            },
        )

    def test_filters_mix(self):
        self._run_test(
            {
                'filters': {
                    'field1': 'banana',
                    'field2': ['lemons', 'blarp'],
                }
            },
            {
                'bool': {
                    'filter': [
                        {
                            'term': {
                                keyword('field1'): 'banana',
                            }
                        },
                        {
                            'term': {
                                keyword('field2'): 'lemons',
                            }
                        },
                        {
                            'term': {
                                keyword('field2'): 'blarp',
                            }
                        },
                    ]
                }
            },
        )



from ckan.plugins.interfaces import Interface
from splitgill.indexing.options import ParsingOptionsBuilder


class IVersionedDatastore(Interface):
    """
    Allow modifying versioned datastore logic.
    """

    def vds_before_search(self, request):
        """
        Called right before the Search object is created and performed. Implementing
        this hook allows modification of the request (including the query etc) before it
        is run. It doesn't matter how the search request is made, this hook will be
        called, so make sure your implementation works with all Query types etc.

        :param request: a SearchRequest instance
        """
        pass

    def vds_after_multi_query(self, response, result):
        pass

    def vds_is_read_only_resource(self, resource_id: str):
        """
        Allows implementors to designate certain resources as read only. This is purely
        a datastore concept and doesn't impact the actual resource from CKAN's point of
        view. This is checked when performing the following actions:

            - vds_data_add
            - vds_data_delete
            - vds_data_sync

        :param resource_id: the resource id to check
        :returns: True if the resource should be treated as read only, False if not
        """
        return False

    def vds_update_options(self, resource_id: str, builder: ParsingOptionsBuilder):
        pass

    def vds_reserve_slugs(self):
        """
        Allows implementors to reserve queries using reserved pretty slugs. Implementors
        should return a dict made up of reserved pretty slugs as keys and then the slug
        parameters as the values. These values should be another dict containing the
        following optional keys:

            - query, a dict query (defaults to {})
            - query_version, the query schema version (defaults to the latest query
                             schema version)
            - version, the version of the data to search at (defaults to None)
            - resource_ids, a list of resource ids to search (defaults to all resource
                            ids)

        If a slug already exists in the database with the same reserved pretty slug and
        the same query parameters then nothing happens.

        If a slug already exists in the database with the same reserved pretty slug but
        a different set of query parameters then a DuplicateSlugException is raised.

        If a slug already exists in the database with the same query parameters but no
        reserved pretty slug then the reserved pretty slug is added to the slug.
        """
        return {}

    def vds_modify_field_groups(self, resource_ids, fields):
        """
        Allows plugins to manipulate the FieldGroups object used to figure out the
        groups that should be returned by the vds_multi_fields action.

        :param resource_ids: a list of resource ids
        :param fields: a FieldGroups object
        """
        pass

    def datastore_before_convert_basic_query(self, basic_query):
        """
        Allows plugins to modify a basic query (probably taken from a URL), e.g. to
        remove custom filters before processing.

        :param basic_query: the query dict to be modified
        :returns: the modified query
        """
        return basic_query

    def datastore_after_convert_basic_query(self, basic_query, multisearch_query):
        """
        Allows plugins to modify a converted query, e.g. to add back in any complex
        custom filters.

        :param basic_query: the original basic query, before it was modified by other
            plugins
        :param multisearch_query: the converted multisearch version of the query
        :returns: the modified multisearch query
        """
        return multisearch_query


class IVersionedDatastoreQuerySchema(Interface):
    def get_query_schemas(self):
        """
        Hook to allow registering custom query schemas.

        :returns: a list of tuples of the format (query schema version, schema object)
            where the query schema version is a string of format v#.#.# and the schema
            object is an instance of ckanext.versioned_datastore.lib.query.Schema
        """
        return []


class IVersionedDatastoreDownloads(Interface):
    def download_modify_notifier_start_templates(
        self, text_template: str, html_template: str
    ):
        """
        Hook allowing other extensions to modify the templates used when sending
        notifications that a download has started. The templates can be modified in
        place or completely replaced.

        :param text_template: the text template string
        :param html_template: the html template string
        :returns: a 2-tuple containing the text template string and the html template
            string
        """
        return text_template, html_template

    def download_modify_notifier_end_templates(
        self, text_template: str, html_template: str
    ):
        """
        Hook allowing other extensions to modify the templates used when sending
        notifications that a download has completed successfully. The templates can be
        modified in place or completely replaced.

        :param text_template: the text template string
        :param html_template: the html template string
        :returns: a 2-tuple containing the text template string and the html template
            string
        """
        return text_template, html_template

    def download_modify_notifier_error_templates(
        self, text_template: str, html_template: str
    ):
        """
        Hook allowing other extensions to modify the templates used when sending
        notifications that a download has failed. The templates can be modified in place
        or completely replaced.

        :param text_template: the text template string
        :param html_template: the html template string
        :returns: a 2-tuple containing the text template string and the html template
            string
        """
        return text_template, html_template

    def download_modify_notifier_template_context(self, request, context):
        """
        Hook allowing other extensions to modify the templating context used to generate
        the download email (both plain text and HTML versions) before it is sent.

        The default context contains:
            - "download_url": the download zip's full URL
            - "site_url": the CKAN site's full URL (this is taken straight from the
                          config)

        :param request: the DownloadRequest object
        :param context: templating context dict
        :returns: context templating dict
        """
        return context

    def download_derivative_generators(self, registered_derivatives=None):
        """
        Extend or modify the list of derivative generators.

        :param registered_derivatives: a dict of existing derivative generator classes,
            returned from previously loaded plugins
        :returns: a dict of loaded derivative generator classes, keyed on the name used
            to specify them in download requests
        """
        return registered_derivatives or {}

    def download_file_servers(self, registered_servers=None):
        """
        Extend or modify the list of file servers.

        :param registered_servers: a dict of existing file server classes, returned from
            previously loaded plugins
        :returns: a dict of loaded file server classes, keyed on the name used to
            specify them in download requests
        """
        return registered_servers or {}

    def download_notifiers(self, registered_notifiers=None):
        """
        Extend or modify the list of download notifiers.

        :param registered_notifiers: a dict of existing notifier classes, returned from
            previously loaded plugins
        :returns: a dict of loaded notifier classes, keyed on the name used to specify
            them in download requests
        """
        return registered_notifiers or {}

    def download_data_transformations(self, registered_transformations=None):
        """
        Extend or modify the list of data transformations.

        :param registered_transformations: a dict of existing data transformations,
            returned from previously loaded plugins
        :returns: a dict of loaded transformations, keyed on the name used to specify
            them in download requests
        """
        return registered_transformations or {}

    def download_before_init(
        self, query_args, derivative_args, server_args, notifier_args
    ):
        """
        Hook allowing other extensions to modify args before any search is run or files
        generated.

        :param query_args: a QueryArgs object
        :param derivative_args: a DerivativeArgs object
        :param server_args: a ServerArgs object
        :param notifier_args: a NotifierArgs object
        :returns: all four args objects (query_args, derivative_args, server_args,
            notifier_args)
        """
        return query_args, derivative_args, server_args, notifier_args

    def download_after_init(self, query):
        """
        Hook notifying that the downloader and associated objects (e.g. the query) have
        been initialised. Does not allow modification; purely for notification purposes.

        :param query: the query for this download
        :returns: None
        """
        return

    def download_modify_manifest(self, manifest, request):
        """
        Hook allowing other extensions to modify the manifest before the download file
        is written. Modifications to the request object are not saved.

        :param manifest: the manifest dict
        :param request: the DownloadRequest object
        :returns: the manifest dict
        """
        return manifest

    def download_modify_eml(self, eml_dict, query):
        """
        Hook allowing other extensions to modify the content of the EML before it's
        transformed into XML and written to file.

        :param eml_dict: the current eml content, as a dict
        :param query: the query for this download
        :returns: the modified eml dict
        """
        return eml_dict

    def download_after_run(self, request):
        """
        Hook notifying that a download has finished (whether failed or completed). Does
        not allow modification; purely for notification purposes.

        :param request: the DownloadRequest object
        :returns: None
        """
        return



from ckantools.decorators import auth


@auth(anon=True)
def vds_options_get(context, data_dict) -> dict:
    # allow access to everyone
    return {'success': True}


@auth('resource_update', {'resource_id': 'id'})
def vds_options_update(context, data_dict) -> dict:
    return {'success': True}
